import asyncio
import typing

async def waitfor(coro: typing.Coroutine):
    """
    Starts a coroutine and shows some animated cursor during the waiting phase
    """
    T = asyncio.ensure_future(coro)
    while not T.done():
        for c in '-\|/':
            print(c, end='\r')
            await asyncio.sleep(0.1)
    print('done')
    return T.result()


def await_coro(coro: typing.Coroutine):
    """
    Starts a coroutine in the standard loop and animates  a cursor
    """
    return asyncio.get_event_loop().run_until_complete(waitfor(coro))



"""
The Bus-System
--------------

Busses are ways for the unilogger to communicate with measurement devices. A Bus can be understood as
a wrapper of the device specific protocol to an universal API. The concrete instance of a bus
is defined with a yaml file - the layout of parameters is heavily depending on the Bus implementation.

In most cases, multiple sensors can be attached to a Bus: Eg. for the SDI12-Bus, a couple of sensors at
different addresses are connected. Each of the sensor can produce multiple values.

But the sensor level is not necessary - in essence a Bus a protocol, that can produce, when read multiple values
with proper meta data.

The general layout of a Bus is:

- Bus -> has multiple Sensors
- Sensor -> has ValueFactorys (see :py:class:`unilogger.bus.base.ValueFactory`)

Every Bus must implement the following methods:

- **__asdict__**: Returns a dictionary that describes a bus with every attached sensor and measured values
- **read_all** (coroutine): Reads all values defined for the bus and returns a list of Values (see: :py:class:`unilogger.bus.Value`)
- **__init__**: The Init function must accept all key words from the defining yaml file

The open_bus function loads the bus description (usally from a yaml file
"""

from .base import open_bus, Value



"""

"""

import os

# Get BOM for excel csv


class Csv:
    def __init__(self, filename, openmode='a'):
        if (not os.path.exists(filename)) or openmode == 'w':
            self.file = open(filename, 'w')
            # Make Excel aware this is UTF-8 coded (BOM)
            self.file.write('\ufeff')
            # Write header
            self.file.write('VariableName, Time, Value, Unit\n')

        else:
            self.file = open(filename, 'a')

    def __call__(self, values, variableprefix=''):
        for v in values:
            self.file.write('"{pf}{v.name}", {time}, {v.value:0.6g}, "{v.unit}"\n'
                            .format(v=v, time=v.time.strftime('%Y-%m-%d %H:%M:%S'), pf=variableprefix))
        self.file.flush()

    def close(self):
        self.file.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()




import pandas as pd

class HDFLogger:
    def __init__(self, filename, hdf_key='data'):
        # self.store = pd.HDFStore(filename)
        self.filename=filename
        self.key = hdf_key

    def __call__(self, values, variableprefix=''):
        df = pd.DataFrame([v.__asdict__() for v in values])
        df.to_hdf(self.filename, self.key, append=True, index=False, data_columns=True)

    def close(self):
        ...
        # self.store.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()






"""

"""
import typing
import sys

import aiohttp
import datetime
from bs4 import BeautifulSoup


from . import base


class AddUPIError(base.BusError):
    pass


class Sensor(base.Sensor):
    """
    A sensor connected to an ADCON network. Any object of the class "DEVICE"
    can be used as a sensor
    """

    def __init__(self, id, name=None, values=None, **kwargs):
        """
        Creates the sensor
        :param id: Id of the DEVICE
        :param name:
        """
        self.id = id
        self.name = name
        if values:
            self.valuefactories = [base.ValueFactory(**v) for v in values]
        else:
            self.valuefactories = []

        self.extradata = kwargs

    def __str__(self):
        return '{} (id={}) ({} values)'.format(self.name, self.id, len(self.valuefactories))
     
    def __repr__(self):
        return 'addupi.Sensor(id={}, name={})'.format(self.id, self.name)

    def __asdict__(self)->dict:
        res = self.extradata.copy()
        res.update(dict(id=self.id, name=self.name, values=[]))
        for vf in self.valuefactories:
            res['values'].append(vf.__asdict__())
        return res


class Bus(base.Bus):
    def __init__(self, url, user=None, password=None, sensors=None, **kwargs):
        self.url = url
        self.session = None
        self.user = user
        self.password = password
        self.extradata = kwargs
        if sensors:
            self.sensors = [Sensor(**s) for s in sensors]
        else:
            self.sensors = []

    def __asdict__(self):
        res = self.extradata.copy()
        res.update(dict(url=self.url, user=self.user, password=self.password))
        res['module'] = __name__
        res['sensors'] = []
        for s in self.sensors:
            res['sensors'].append(s.__asdict__())
        return res

    def __str__(self):
        return 'ADCONhost on {} with {} sensors'.format(self.url, len(self.sensors))

    async def __aenter__(self):
        await self.login()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.logout()

    def makesensor(self, id: str, name=None, values=None, **kwargs):
        """
        Create a sensor on this bus
        :param id: id of the DEVICE node (str)
        :param name: Name of the DEVICE node
        :param values: List of dict, describing valuefactories
        :param kwargs: Extra data to describe the sensor
        :return: The created AddUPISensor
        """
        s = Sensor(self, id, name)
        if values:
            errors = []
            for v in values:
                try:
                    s.valuefactories[v['id']] = base.ValueFactory(**v)
                except Exception as e:
                    errors.append(e)
                else:
                    errors.append(None)
            # TODO: Do something about the errors
        self.sensors.append(s)
        return s

    async def read(self, **params) ->(BeautifulSoup, str):
        """
        Reads from self.url using the params and returns the souped response
        :param params: A dictionary of parameters
        :return: a BeautifulSoup of the response
        """
        if self.session and 'session-id' not in params:
            params['session-id'] = self.session
        async with aiohttp.ClientSession() as session:
            # GET response to function
            async with session.get(self.url, params=params) as r:
                # check response

                # status check
                if r.status // 100 > 2:
                    raise AddUPIError('AddUPI connection failed, got status {} from {}'.format(r.status, r.url))
                # get text
                text = await r.text()
                # parse text
                try:
                    soup = BeautifulSoup(text, 'xml')
                except Exception as e:
                    raise AddUPIError('AddUPI connection failed, could not parse response from {}\nResponse:\n{}'
                                      .format(r.url, text)) from e
                # check for response
                if not soup.response:
                    raise AddUPIError('AddUPI connection failed, no response tag from {}, got instead:\n{}'
                                      .format(r.url, soup.prettify()))
                # check for error
                if error := soup.response.find('error', recursive=False):
                    raise AddUPIError('AddUPI connection failed got Error code {code}: {msg} on {url}'
                                      .format(url=r.url, **error.attrs))
                return soup, r.url

    async def login(self, timeout=None):
        """
        Does a login to the ADCON server using the configured username and password
        :param timeout: A timeout for the session in seconds
        :return: A new session id
        """
        params = dict(function='login', user=self.user, passwd=self.password)
        if timeout:
            params['timeout'] = timeout
        soup, url = await self.read(**params)
        self.session = str(soup.response.result.text.strip())
        if not self.session or self.session=='None':
            raise AddUPIError('Tried to login but got not a session id from {} Response:\n{}'.format(url, soup))
        return self.session

    async def logout(self):
        """
        Finishes the current session
        """
        if not self.session:
            raise AddUPIError("AddUPI is not connected, can't logout")
        params = {'function': 'logout', 'session-id': self.session}
        await self.read(**params)
        self.session = None

    async def configsensor(self, sensorid) -> Sensor:
        """
        Gets the configuration of a sensor from the getconfig function of the AddUPI protocol
        :param sensorid: Id of the sensor
        :return: The configured sensor
        """
        if not self.session:
            raise AddUPIError('addupi.Bus.configsensor: No log in session available')
        soup, url = await self.read(function='getconfig', id=sensorid, depth=1)
        sensor = soup.response.node
        if not sensor or sensor['id'] != str(sensorid):
            raise AddUPIError('AddUPI asked for sensor #{id} with {url} but got:\n{text}'
                              .format(id=sensorid, url=url, text=soup.prettify()))
        if sensor['class'] != 'DEVICE':
            raise AddUPIError('{} has not the class DEVICE'.format(sensor))

        asensor = Sensor(sensorid, sensor['name'])
        for n in sensor.nodes.find_all('node'):
            if n['class'] == 'TAG':
                asensor.valuefactories.append(base.ValueFactory(**n.attrs))
        self.sensors.append(asensor)
        return asensor

    async def read_all(self) -> typing.List[base.Value]:
        if not self.session:
            raise AddUPIError('addupi.Bus.read_all: No log in session available')
        values = []
        for sensor in self.sensors:
            s_values = await self.readsensor(sensor)
            values.extend(s_values)
        return values

    async def readsensor(self, sensor: Sensor, fromdate=None, slots=None) -> typing.List[base.Value]:
        """
        Reads a sensor
        :param sensor: AddUPISensor to read
        :return: list of values (Value)
        """
        if not self.session:
            raise AddUPIError('addupi.Bus.read_all: No log in session available')

        params = {'function': 'getdata',
                  'id': sensor.id}
        if fromdate:
            params['date'] = fromdate.strftime('%Y%m%dT%H:%M:%S')
            params['slots'] = slots or 1000

        soup, url = await self.read(**params)
        # Check returned node_id
        nodes = soup.response.find_all('node')
        # Make an empty value list
        values = []
        # Make a dict to relate id's with valuefactory numbers
        vfdict = dict((str(vf.id), i) for i, vf in enumerate(sensor.valuefactories))
        for node in nodes:
            # Get first value with absolute date
            node_id = node['id']
            if node_id in vfdict:
                if error := node.find('error', recursive=False):
                    code = error.attrs.get('code')
                    msg = error.attrs.get('msg')
                    print(f'{datetime.datetime.now()}: Addupi warning at node {node_id}: #{code}: {msg} ({url})')
                else:
                    for v_elem in node('v', recursive=False):
                        t_str = v_elem['t']
                        try:
                            t = datetime.datetime.strptime(t_str, '%Y%m%dT%H:%M:%S')
                        except ValueError:
                            if v_elem['t'].startswith('+'):
                                t += datetime.timedelta(seconds=int(v_elem['t']))
                            else:
                                raise
                        v = float(v_elem.string)
                        # Check for relative time
                        i = vfdict[node_id]
                        values.append(sensor.valuefactories[i](v, t))


        return values



from datetime import datetime
import typing
import importlib
import yaml


import logging
logger = logging.getLogger(__name__)

class BusError(Exception):
    ...


class ScaleFunction:
    """
    A user defined function to scale or translate a measured value into something meaningful
    """
    def __init__(self, code: str, testvalue=None):
        if 'x' not in code:
            raise ValueError('Function code {} must include the variable x'.format(code))
        self.__code = code
        try:
            import math
            self.function = eval('lambda x:' + code, vars(math))
            if testvalue is not None:
                self.function(testvalue)
        except Exception as e:
            raise ValueError('f(1) = {} is not a valid python expression, got error: {}'
                             .format(code.replace('x', '1'), repr(e)))

    def __call__(self, x: float):
        return self.function(x)

    def __repr__(self):
        return "ScaleFunction('{}')".format(self.__code)

    def __str__(self):
        return self.__code


class Value:
    """
    A measured value with metadata
    """
    def __init__(self, value, time=None, name=None, unit=None, **kwargs):
        """
        Creates the value with meta data
        :param value: The value (a float)
        :param time: Time of measurement
        :param name: Name of measurement
        :param datasetid: Target dataset id in external database (eg- Schwingbach)
        :param kwargs: Additional meta data
        """
        self.value = value
        self.name = name
        self.time = time
        self.unit = unit
        self.extradata = kwargs

    def __str__(self):
        res = ''
        if self.name:
            res += self.name + '='
        res += '{:0.6g}'.format(self.value)
        if hasattr(self, 'unit') and self.unit:
            res += ' ' + str(self.unit)
        if self.time:
            res += self.time.strftime(' (%d.%m.%Y %H:%M:%S)')
        return res

    def __repr__(self):
        res = 'name={name!r}, time={time!r}, value={value:0.4g}'.format(**vars(self))
        if self.extradata:
            res += ', ' + ', '.join('{!s}={!r}'.format(*it) for it in self.extradata.items())
        return 'Value({})'.format(res)

    def __getattr__(self, item):
        if item in self.extradata:
            return self.extradata[item]
        else:
            raise AttributeError(f'{item} not an attribute of {self}')

    def __getitem__(self, item):
        if item in self.extradata:
            return self.extradata[item]
        else:
            raise KeyError(f'{item} not a data field of {self}')

    def __asdict__(self):
        """
        :return: The Value as a dictionary
        """
        res = self.extradata.copy()
        res.update(dict(name=self.name, value=self.value, time=self.time.isoformat()))
        return res


class ValueFactory:
    """
    A value factory is a kind of template to create a value with all needed metadata
    """

    def __init__(self, name: str = None, unit: str = None, scalefunction: typing.Union[str, ScaleFunction] = None, **kwargs):
        """

        :param name: Name of the measured item
        :param unit: Unit of the measured item
        :param scalefunction: A valid python expression to transform the raw data into the output value
        :param kwargs: Extraarguments passed to the Value
        """
        self.name = name
        self.unit = unit
        if isinstance(scalefunction, ScaleFunction):
            self.scalefunction = scalefunction
        elif scalefunction:
            self.scalefunction = ScaleFunction(scalefunction)
        else:
            self.scalefunction = None
        self.extradata = kwargs

    def __asdict__(self) -> dict:
        """
        Creates a dict describing te
        :return:
        """
        res = self.extradata.copy()
        res.update(dict(name=self.name, unit=self.unit, scalefunction=str(self.scalefunction) if self.scalefunction else None))
        return res

    def __call__(self, value: float, time=None, **kwargs) -> Value:
        """
        Creates a Value from a float
        :param value: the measured value
        :param time: A datetime of the measurement, if None, utcnow() is used
        :param kwargs: Extra meta data to be stored with the value
        :return: Value
        """
        if self.scalefunction:
            value = self.scalefunction(value)
        time = time or datetime.utcnow()
        data = self.extradata.copy()
        data.update(kwargs)
        return Value(value, time, self.name, unit=self.unit, **data)

    def __getattr__(self, item):
        if item in self.extradata:
            return self.extradata[item]
        else:
            raise AttributeError(f'{item} not an attribute of {self}')

    def __getitem__(self, item):
        if item in self.extradata:
            return self.extradata[item]
        else:
            raise KeyError(f'{item} not a data field of {self}')

    def __repr__(self):
        if self.scalefunction:
            name = str(self.scalefunction).replace('x', self.name)
        else:
            name = self.name
        return '{name} [{self.unit}]'.format(name=name, self=self)

class Sensor:
    """
    Base class for a sensor on a bus
    """
    valuefactories: typing.List[ValueFactory] = None


class Bus:
    """
    Base class for a generic bus system.
    General hierachy of the logger system:
    Bus (eg. SDI12)
    -->has sensors (eg. VAISALA @ address 0)
       -->has valuefactories (eg. Air Temp)
    """
    sensors: typing.List[Sensor] = None

    def __asdict__(self):
        raise NotImplementedError

    async def readsensor(self, sensor):
        raise NotImplementedError

    async def read_all(self) -> typing.List[Value]:
        """
        Reads all sensors
        :return: A List of Values
        """
        raise NotImplementedError


    @classmethod
    def from_dict(cls, data: dict, **kwargs):
        """
        Loads a bus from a dictionary (eg. given by JSON or YAML).
        The bus type is loaded from the module key of the dictionary
        :param data: The dictionary describing the bus
        :return: module.Bus, where module is the module given in data['module']
        """
        data.update(kwargs)
        if 'module' not in data:
            raise KeyError('To create a generic logger.Bus you need to provide the module')
        modname = data.pop('module')
        try:
            module = importlib.import_module(modname)
        except Exception:
            raise ValueError('Module "{}" not found, check configuration'.format(modname))
        else:
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, cls):
                    logger.debug(f'Found Bus:{obj.__name__}')
                    return obj(**data)
            raise KeyError('Module "{}" exists, but has no Bus class'.format(modname))

    @classmethod
    def from_file(cls, busfile, **kwargs):
        """
        Loads a Bus from a description file
        :param busfile: Filename of the yaml description of the bus
        :return: the bus
        """
        data = yaml.safe_load(busfile)
        data.update(kwargs)
        return cls.from_dict(data)

    def to_file(self, busfile):
        """
        Saves the current bus description as a yaml file
        :param busfile: Filename to save. For further use, the file should match to: preferences/*.bus.yaml
        """
        self.to_stream(open(busfile, 'w'))

    def to_stream(self, stream):
        yaml.safe_dump(
            self.__asdict__(),
            stream,
            default_flow_style=False,
            sort_keys=False,
        )


__OpenTypes = typing.Optional[typing.Union[typing.Mapping, typing.IO, str]]


def open_bus(source: __OpenTypes = None, **kwargs) -> Bus:
    """
    Loads a Bus from a dict, a stream or a filename
    :param source: Either a dict, a file stream, a filename or a string containing a yaml-document
    :return: The created Bus
    """
    import os
    if isinstance(source, typing.Mapping):
        return Bus.from_dict(source, **kwargs)
    elif isinstance(source, typing.IO):
        return Bus.from_file(source, **kwargs)
    elif isinstance(source, str) and os.path.exists(source):
        with open(source) as f:
            return Bus.from_file(f, **kwargs)
    elif isinstance(source, str):
        try:
            d = yaml.safe_load(source)
        except:
            raise BusError(f'open_bus: Source does not contain a Bus-Description: {source}')
        else:
            return Bus.from_dict(d, **kwargs)
    elif source is None:
        return Bus.from_dict(kwargs)
    else:
        raise BusError(f'open_bus: Cannot handle source: {source}')



import asyncio
import time

class TimeRaster:
    def __init__(self, seconds, offset=0):
        self.seconds = seconds
        self.offset = offset % seconds

    def due_in(self):
        now = time.time()
        return self.seconds - (now - self.offset) % self.seconds

    def progress(self):
        return 1 - self.due_in() / self.seconds

    def __str__(self):
        return 'every {}, {}, {} ... seconds'.format(*[s*self.seconds+self.offset for s in range(3)])


every10s = TimeRaster(10)
every30s = TimeRaster(30)
every1minute = TimeRaster(60)
every5minutes = TimeRaster(5 * 60)
every15minutes = TimeRaster(10 * 60)
every30minutes = TimeRaster(30 * 60)
every1hour = TimeRaster(60 * 60)


class ReadTask(asyncio.Future):
    async def read(self):
        try:
            res = await self.bus.read_all()
        except Exception as e:
            self.set_exception(e)
        else:
            self.set_result(res)

    def result_or_exception(self):
        if self.exception():
            return self.exception()
        else:
            return self.result()

    def __init__(self, bus):
        super().__init__()
        self.bus = bus
        self.task = asyncio.ensure_future(self.read())

    def __repr__(self):
        return 'T({})'.format(self.bus)


class Schedule:
    """
    Schedules the read out of the logger busses it relates to timerasters
    """

    def __init__(self, seconds, busses=None, offset=0, onread=None, *, loop=None):
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop
        self.busses = busses or []
        self.raster = TimeRaster(seconds, offset)
        print('Schedule: 1st reading in {:0.1f}s'.format(self.raster.due_in()))
        self.onread = onread
        self.task = self.loop.call_later(self.raster.due_in(), asyncio.ensure_future, self.read())

    def __str__(self):
        return ','.join(str(bus) for bus in self.busses) + ' (' + str(self.raster) + ')'

    def cancel(self):
        self.task.cancel()
        print(time.ctime(), 'Cancelled:', self)

    async def read(self):
        # loop.call_later is not precise. make sure the time to run has really come
        if self.raster.due_in() < self.raster.seconds * 0.5:
            await asyncio.sleep(self.raster.due_in())
        # Read the busses
        pending = [ReadTask(bus) for bus in self.busses]
        while pending:
            done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED, timeout=self.raster.due_in())
            d: ReadTask
            if self.onread:
                self.onread([(d.bus, d.result_or_exception()) for d in done])

        self.task = self.loop.call_later(self.raster.due_in(), asyncio.ensure_future, self.read())


import datetime


async def read_at(t: datetime.datetime, *busses):
    """
    Waits until t and reads all sensors of the given busses
    :param t:
    :param busses:
    :return:
    """
    sleeptime = (t - datetime.datetime.now()).total_seconds()
    await asyncio.sleep(sleeptime)
    tasks = [b.read_all() for b in busses]
    value_lists = await asyncio.gather(*tasks)
    return sum(value_lists, start=[])



if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    schedules = [Schedule(5,'abc'), Schedule(5, 'xyz', offset=3)]
    loop.call_later(300, schedules[0].cancel)
    loop.run_forever()



from setuptools import setup, find_packages

setup(
    name='unilogger',
    version='2022.9.13',
    packages=find_packages(),
    author='Philipp Kraft',
    author_email='philipp.kraft@umwelt.uni-giessen.de',
    description='A unified wrapper to log data from various sensor busses',
    install_requires=[
        'umodbus>1.0', 'aiohttp>3.7', 'BeautifulSoup4>4.9',
        'aioserial>=1.3.0', 'pyyaml>=6.0', 'lxml', 'pyserial>=3.5',
        'asteval>=0.9', 'pandas>=1.4', 'tables>=3.6', 'openpyxl'
    ],
    python_requires='>=3.8'
)


"""
Logger bus wrapper for SDI12 devices

SDI12 specification: http://www.sdi-12.org/current%20specification/SDI-12_version-1_4-August-10-2016.pdf
"""
from . import base
import aioserial
import asyncio
import datetime
from attrdict import AttrDict
from typing import List
import contextlib

import warnings

import logging
logger = logging.getLogger(__name__)

def encode(string: str) -> bytes:
    """
    Encodes a unicode string to a bytesting with UTF-8
    :param string:
    :return:
    """
    return string.encode()


def decode(bytestring: bytes) -> str:
    """
    Decodes a bytestring to unicode assuming UTF-8
    :param bytestring:
    :return:
    """
    return bytestring.decode()


class parser:
    """
    Helper functions to parse SDI12 responses
    """
    channels = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

    @staticmethod
    def M(response: str):
        """
        Parses the response to a C or M command
        :return: seconds to result, number of values
        """
        try:
            return int(response[1:4]), int(response[4:])
        except (ValueError, TypeError):
            raise RuntimeError('Expected numbers but got {}'.format(response))

    @staticmethod
    def D(response: str):
        """
        Parses the response to an SDI12 D command and yields the values
        :param response: Typical response to D command is a+xxx.x+xxx.xxx-xxx
        :return: List of float values
        """
        # Add spaces between the numbers, split the response and skip the address
        values = response.replace('+', ' +').replace('-', ' -').split()[1:]
        for value in values:
            try:
                v = float(value)
                yield v
            except (ValueError, TypeError):
                yield None


class SDI12port(aioserial.AioSerial):
    def __init__(self, port, **kwargs):
        super().__init__(port, **kwargs)
        self.lock = asyncio.Lock()

    async def __call__(self, cmd: str) -> str:
        """
        Writes string cmd to serial port and return the response as string
        :param cmd:
        :return:
        """
        with await self.lock:
            await self.write_async(encode(cmd + '\n'))
            # SDI 12 is slow, let some time pass until looking for a result
            await asyncio.sleep(0.1)
            line = await self.readline_async()
            return decode(line)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Sensor(base.Sensor):
    """
    Wraps a sensor on a SDI12-Bus connected over the Arduino based
    SDI12 connector
    """

    def __init__(self, channel, name, values, **kwargs):
        """
        Creates a new SDI12device from the response to the aI! string
        :bus: the SDI12bus this device belongs to.
        :info: A string line response from the SDI12 aI! command
        :config: A list of dictionaries
        """
        self.channel = channel
        self.name = name
        self.extradata = kwargs
        self.valuefactories: List[base.ValueFactory] = [
            base.ValueFactory(**v)
            for v in values
        ]
        for vf in self.valuefactories:
            if not vf.name.startswith(self.name):
                vf.name = self.name + '.' + vf.name

    def __str__(self):
        return ('SDI12 device on {o.channel} is {o.name} ({c} values)'
                .format(o=self, c=len(self.valuefactories)))

    async def do_measurement(self, serial: SDI12port) -> asyncio.Task:
        """
        Starts an asynchronous measurement of the device and returns
        a task to read the measurement later with :meth:`unilogger.bus.sdi12.Sensor.read_measurement`

        SDI-Command sent: aC!
        SDI-Command of the Task: aD0!, if necessary also aD1! etc.

        :param serial: The SDI12 port
        :return: asyncio.Task to write aDX! commands
        """
        response = await serial(self.channel + 'C!\r\n')
        # get wait time and number of values for this device
        wait, nvalues = parser.M(response)
        logging.debug(f'SDI12.{self.channel}C: {nvalues} values in {wait}s')
        return asyncio.create_task(self.read_measurement(serial, wait, nvalues))

    async def read_measurement(self, serial: SDI12port, wait: float, nvalues: int) -> List[base.Value]:
        """
        Wait the wait time and performs the D0 (get data) command
        If necessary it calls also the D1, D2 etc commands
        :param serial: Write coroutine
        :param wait: Wait time in seconds
        :param nvalues: Number of values
        :return: A list of augmented Values
        """
        tstart = datetime.datetime.utcnow()
        await asyncio.sleep(wait)
        # Get the values, probably spread over several D commands
        floatvalues = []
        d = 0  # The D command number
        # repeat the D commands as necessary
        while len(floatvalues) < nvalues:
            response = await serial('{}D{}!\r\n'.format(self.channel, d))
            floatvalues.extend(parser.D(response))
            d += 1
        # Augment & transform the values
        return [factory(x, time=tstart) for
                factory, x in
                zip(self.valuefactories, floatvalues)]

    def __asdict__(self)->dict:
        res = self.extradata.copy()
        res.update(dict(channel=self.channel, name=self.name, values=[]))
        for vf in self.valuefactories:
            res['values'].append(vf.__asdict__())
        return res


class SDI12Bus(base.Bus):
    """
    Talks to configured SDI12-Sensors at this SDI12-Bus
    """

    def __init__(self, port=None, sensors=None, **kwargs):
        kwargs.pop('module', None)

        self.port = port
        self.baudrate = 9600
        self.timeout = 0.5
        self.extradata = kwargs

        self.port = port

        if sensors:
            self.sensors = [Sensor(**sensor_data) for sensor_data in sensors]

    @contextlib.contextmanager
    def open(self) -> SDI12port:
        with SDI12port(self.port, timeout=self.timeout, baudrate=self.baudrate) as port:
            yield port

    def __repr__(self):
        return 'sdi12.Bus(port={})'.format(self.port)

    async def change_address(self, old_adress, new_adress):
        with self.open() as serial:
            await serial(old_adress + 'A' + new_adress + '!')

    def makesensor(self, channel, name=None, values=None, **kwargs) -> Sensor:
        """
        Create a sensor on this bus
        :param channel: SDI12 channel ('0'..'9', 'a'..'z','A'..'Z')
        :param name: Name of the sensor
        :param values: List of dict, describing valuefactories
        :param kwargs: Extra data to describe the sensor
        :return: The created SDI12Sensor
        """
        s = Sensor(channel, name, **kwargs)
        if values:
            errors = []
            for v in values:
                try:
                    s.valuefactories.append(base.ValueFactory(**v))
                except Exception as e:
                    logging.warning(f'Problem with')
        return s

    def __asdict__(self):
        res = self.extradata.copy()
        res.update(dict(port=self.port))
        res['module'] = __name__
        res['sensors'] = []
        for s in self.sensors:
            res['sensors'].append(s.__asdict__())
        return res


    async def scanchannel(self, serial: SDI12port, channel: str) -> Sensor:
        """
        Reads on the specified SDI12 channel to create a sdi12.Sensor

        :param serial: SDI12port instance
        :param channel: SDI12 channel ('0'..'9', 'a'..'z','A'..'Z')
        :return:
        """

        logging.info('Channel: ', channel, end='\n')

        # Check channel with Acknowledge Active Command
        response = await serial(channel + '!\r\n')
        if response.strip() == channel:
            name = (await serial(channel + 'I!\r\n')).strip()
            logging.info(f'{channel} -> {name}')
            # Get number of values with C command to add valuefactories
            response = await serial(channel + 'C!\r\n')
            wait, nvalues = parser.M(response)
            # Get valuefactory stubs
            valuefactories = [
                base.ValueFactory(name='Value_{:02d}'.format(i), id=i)
                for i in range(nvalues)
            ]
            logging.info(f'     {nvalues} values in {wait} seconds')
            return self.makesensor(channel, name, valuefactories, scantime=wait)

        elif response.strip():
            warnings.warn('Queried channel {} but got {} as answer'.format(channel, response.strip()))

    async def scanbus(self, channels=parser.channels) -> List[Sensor]:
        """
        Scans the SDI12 bus on the channels and creates sensor stubs
        :param channels: A string of channel characters
        :return: List of detected sensors

        Usage:
        >>> bus = Bus('COM1')
        >>> bus.sensors = asyncio.run(bus.scanbus('01234'))
        >>> values = asyncio.run(bus.read_all())
        >>> for v in values:
        >>>     print(v)
        """
        with SDI12port(self.port, timeout=self.timeout, baudrate=self.baudrate) as serial:
            sensors = []
            for c in channels:
                sensor = await self.scanchannel(serial, c)
                if sensor:
                    sensors.append(sensor)
            return sensors

    async def read_all(self):
        """
        Reads all sensors concurrently
        :return:
        """
        return await self.readsensor(*self.sensors)


    async def readsensor(self, *sensors):
        """
        Reads the given sensor(s)
        :param sensors: SDI12 sensor objects
        :return: List of Value objects created by the sensors valuefactories
        """
        with self.open() as serial:
            # Storages for wait and value numbers
            tasklist = [(await s.do_measurement(serial)) for s in sensors]

            # Get the data from the sensors
            values = await asyncio.gather(*tasklist)
            return sum(values, [])  # slow but acceptable way to flatten list of lists





import asyncio
import json

from umodbus.client import tcp

from . import base


class Sensor(base.Sensor):
    def __init__(self, mdbunit, startaddress, length, name=None, values=None, **kwargs):
        self.mdbunit = mdbunit
        self.name = name
        self.extradata = kwargs
        self.startaddress = startaddress
        self.length = length
        if values:
            self.valuefactories = [base.ValueFactory(**v) for v in values]
        else:
            self.valuefactories = []

    def __repr__(self):
        return ('{}.Sensor(mdbunit={self.mdbunit}, startaddress={self.startaddress}, ' +
                'length={self.length}, name={self.name})').format(__name__, self=self)

    def __str__(self):
        return '{} (unit={}) ({} values)'.format(self.name, self.mdbunit, len(self.valuefactories))

    def __asdict__(self)->dict:
        res = self.extradata.copy()
        res.update(dict(mdbunit=self.mdbunit, name=self.name,
                        startaddress=self.startaddress, length=self.length,
                        values=[]))
        for vf in self.valuefactories:
            res['values'].append(vf.__asdict__())
        return res


async def send_message(adu, reader, writer):
    """
    Sends a adu message to the tcp client writer/reader pair
     - reader and write are created with asyncio.open_connection
     - the adu is a mdb message
    """
    writer.write(adu)
    await writer.drain()
    response = await reader.read(1024)
    return tcp.parse_response_adu(response, adu)


class Bus(base.Bus):
    def __init__(self, host, port=512, sensors=None, **kwargs):
        self.client = (host, port)
        self.extradata = kwargs
        if sensors:
            self.sensors = [Sensor(**s) for s in sensors]
        else:
            self.sensors = []

    def __asdict__(self):
        res = self.extradata.copy()
        res.update(dict(host=self.client[0], port=self.client[1]))
        res['module'] = __name__
        res['sensors'] = []
        for s in self.sensors:
            res['sensors'].append(s.__asdict__())
        return res

    def __str__(self):
        return 'ModbusTCP client on {} with {} sensors'.format(self.client[0], len(self.sensors))

    async def readsensor(self, sensor: Sensor):
        """
        Reads a single sensor
        """
        r, w = await asyncio.open_connection(*self.client)
        msg = tcp.read_holding_registers(sensor.mdbunit, sensor.startaddress, sensor.length)
        resp = await send_message(msg, r, w)
        w.close()
        
        return [vf(resp[vf.id-sensor.startaddress]) for vf in sensor.valuefactories]

    async def read_all(self):
        values = []
        for sensor in self.sensors:
            s_values = await self.readsensor(sensor)
            values.extend(s_values)
        return values

    def sensors_from_csv(self, filename):
        """
        Creates sensors from a csv file with the columns:
        unit (slave_id), name, description, address, typename, scalefactor, unit
        :param filename: tge filename
        :return:
        """
        slave_id = None
        with open(filename) as f:
            for line in f:
                if line.strip().startswith('#'):
                    continue
                ls = line.split(',', 6)
                if ls[0].strip() != str(slave_id):
                    try:
                        slave_id = int(ls[0].strip())
                    except (ValueError, TypeError):
                        continue
                    address = int(ls[3])
                    self.sensors.append(Sensor(slave_id, address, 0))
                s = self.sensors[-1]
                # Add a value to the sensor
                s.length += 1
                # 1 name, 2 description, 3 address, 4 typename, 5 scalefactor, 6 unit
                name = ls[1].strip()
                unit = ls[6].strip()
                id = int(ls[3].strip())
                # Get the scale function. When the unit is wrapped by braces, assume a lookup dict in the unit
                if unit.strip().startswith('{') and unit.strip().endswith('}'):
                    scalefunction = base.ScaleFunction(unit.strip() + '[x]', testvalue=None)
                    unit = ''
                elif ls[5]:
                    scalefunction = base.ScaleFunction('x / ' + ls[5].strip())
                else:
                    scalefunction = None

                vf = base.ValueFactory(name=name, unit=unit,
                                       scalefunction=scalefunction,
                                       id=id, description=ls[2].strip())
                s.valuefactories.append(vf)
